@@ -12,52 +12,67 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import json
 import logging
 import os
-from typing import Iterator, List, Optional, TypedDict, Union
+from typing import Iterable, Iterator, List, Optional, Union
 
+from ....core.scheduler import InferenceRequest
+from ....device_utils import (
+    get_device_preferred_dtype,
+    gpu_count,
+    is_hf_accelerate_supported,
+)
 from ....types import (
     ChatCompletion,
     ChatCompletionChunk,
     ChatCompletionMessage,
     Completion,
     CompletionChunk,
+    CreateCompletionTorch,
     Embedding,
     EmbeddingData,
     EmbeddingUsage,
+    LoRA,
+    PytorchGenerateConfig,
+    PytorchModelConfig,
 )
+from ...utils import select_device
 from ..core import LLM
 from ..llm_family import LLMFamilyV1, LLMSpecV1
 from ..utils import ChatModelMixin
+from .utils import get_context_length, get_max_src_len
 
 logger = logging.getLogger(__name__)
 
-
-class PytorchGenerateConfig(TypedDict, total=False):
-    temperature: float
-    repetition_penalty: float
-    top_p: float
-    top_k: int
-    stream: bool
-    max_tokens: int
-    echo: bool
-    stop: Optional[Union[str, List[str]]]
-    stop_token_ids: Optional[Union[int, List[int]]]
-    stream_interval: int
-    model: Optional[str]
-
-
-class PytorchModelConfig(TypedDict, total=False):
-    revision: Optional[str]
-    device: str
-    gpus: Optional[str]
-    num_gpus: int
-    max_gpu_memory: str
-    gptq_ckpt: Optional[str]
-    gptq_wbits: int
-    gptq_groupsize: int
-    gptq_act_order: bool
-    trust_remote_code: bool
+NON_DEFAULT_MODEL_LIST: List[str] = [
+    "baichuan-chat",
+    "baichuan-2-chat",
+    "vicuna-v1.3",
+    "falcon",
+    "falcon-instruct",
+    "chatglm",
+    "chatglm2",
+    "chatglm2-32k",
+    "chatglm2-128k",
+    "chatglm3",
+    "chatglm3-32k",
+    "chatglm3-128k",
+    "glm4-chat",
+    "glm4-chat-1m",
+    "llama-2",
+    "llama-2-chat",
+    "internlm2-chat",
+    "qwen-vl-chat",
+    "OmniLMM",
+    "yi-vl-chat",
+    "deepseek-vl-chat",
+    "internvl-chat",
+    "mini-internvl-chat",
+    "cogvlm2",
+    "MiniCPM-Llama3-V-2_5",
+    "glm-4v",
+]
 
 
 class PytorchModel(LLM):
@@ -69,12 +84,14 @@ class PytorchModel(LLM):
         quantization: str,
         model_path: str,
         pytorch_model_config: Optional[PytorchModelConfig] = None,
+        peft_model: Optional[List[LoRA]] = None,
     ):
         super().__init__(model_uid, model_family, model_spec, quantization, model_path)
         self._use_fast_tokenizer = True
         self._pytorch_model_config: PytorchModelConfig = self._sanitize_model_config(
             pytorch_model_config
         )
+        self._peft_model = peft_model
 
     def _sanitize_model_config(
         self, pytorch_model_config: Optional[PytorchModelConfig]
@@ -88,22 +105,24 @@ class PytorchModel(LLM):
         pytorch_model_config.setdefault("gptq_act_order", False)
         pytorch_model_config.setdefault("device", "auto")
         pytorch_model_config.setdefault("trust_remote_code", True)
+        pytorch_model_config.setdefault("max_num_seqs", 16)
         return pytorch_model_config
 
     def _sanitize_generate_config(
         self,
-        pytorch_generate_config: Optional[PytorchGenerateConfig],
+        generate_config: Optional[PytorchGenerateConfig],
     ) -> PytorchGenerateConfig:
-        if pytorch_generate_config is None:
-            pytorch_generate_config = PytorchGenerateConfig()
-        pytorch_generate_config.setdefault("temperature", 0.7)
-        pytorch_generate_config.setdefault("repetition_penalty", 1.0)
-        pytorch_generate_config.setdefault("max_tokens", 512)
-        pytorch_generate_config.setdefault("stream_interval", 2)
-        pytorch_generate_config["model"] = self.model_uid
-        return pytorch_generate_config
+        if generate_config is None:
+            generate_config = PytorchGenerateConfig(**CreateCompletionTorch().dict())
+        else:
+            # Validate generate_config and fill default values to the generate config.
+            generate_config = PytorchGenerateConfig(
+                **CreateCompletionTorch(**generate_config).dict()
+            )
+        generate_config["model"] = self.model_uid
+        return generate_config
 
-    def _load_model(self, kwargs: dict):
+    def _load_model(self, **kwargs):
         try:
             from transformers import AutoModelForCausalLM, AutoTokenizer
         except ImportError:
@@ -128,6 +147,30 @@ class PytorchModel(LLM):
         )
         return model, tokenizer
 
+    def _apply_lora(self):
+        if self._peft_model is not None:
+            try:
+                from peft import PeftModel
+            except ImportError:
+                raise ImportError(
+                    f"Failed to import 'PeftModel' from 'peft'. Please make sure 'peft' is installed.\n\n"
+                )
+
+            for i, peft_model in enumerate(self._peft_model):
+                if i == 0:
+                    self._model = PeftModel.from_pretrained(
+                        self._model,
+                        peft_model.local_path,
+                        adapter_name=peft_model.lora_name,
+                    )
+                else:
+                    self._model.load_adapter(
+                        peft_model.local_path, adapter_name=peft_model.lora_name
+                    )
+                logger.info(
+                    f"PEFT adaptor '{peft_model.lora_name}' successfully loaded for model '{self.model_uid}'."
+                )
+
     def load(self):
         try:
             import torch
@@ -137,23 +180,18 @@ class PytorchModel(LLM):
             )
         from .compression import load_compress_model
 
-        cuda_visible_devices_env = os.getenv("CUDA_VISIBLE_DEVICES", None)
-        cuda_visible_devices = (
-            cuda_visible_devices_env.split(",") if cuda_visible_devices_env else []
-        )
-
         quantization = self.quantization
-        num_gpus = len(cuda_visible_devices) if cuda_visible_devices_env != "-1" else 0
+        num_gpus = gpu_count()
         device = self._pytorch_model_config.get("device", "auto")
-        self._pytorch_model_config["device"] = self._select_device(device)
+        self._pytorch_model_config["device"] = select_device(device)
         self._device = self._pytorch_model_config["device"]
 
-        if self._device == "cpu":
-            kwargs = {"torch_dtype": torch.float32}
-        elif self._device == "cuda":
-            kwargs = {"torch_dtype": torch.float16}
-        elif self._device == "mps":
-            kwargs = {"torch_dtype": torch.float16}
+        kwargs = {}
+
+        dtype = get_device_preferred_dtype(self._device)
+
+        if dtype is not None:
+            kwargs["torch_dtype"] = dtype
         else:
             raise ValueError(f"Device {self._device} is not supported in temporary")
 
@@ -163,10 +201,26 @@ class PytorchModel(LLM):
         kwargs["trust_remote_code"] = self._pytorch_model_config.get(
             "trust_remote_code"
         )
+        model_format = self.model_spec.model_format
 
-        if quantization != "none":
+        is_device_map_auto = False
+
+        # This is required for Intel GPU to actually work with accelerate device_map until
+        # https://github.com/intel/intel-extension-for-pytorch/issues/522
+        # is resolved
+        max_memory_env = os.getenv("ACCELERATE_MAX_MEMORY", None)
+
+        if max_memory_env is not None:
+            max_memory_raw = json.loads(max_memory_env)
+            max_memory = {
+                int(k) if k.isdigit() else k: max_memory_raw[k] for k in max_memory_raw
+            }
+            kwargs["max_memory"] = max_memory
+
+        if quantization != "none" and model_format == "pytorch":
             if self._device == "cuda" and self._is_linux():
                 kwargs["device_map"] = "auto"
+                is_device_map_auto = True
                 if quantization == "4-bit":
                     kwargs["load_in_4bit"] = True
                     kwargs["bnb_4bit_compute_dtype"] = torch.float16
@@ -200,58 +254,25 @@ class PytorchModel(LLM):
                     logger.debug(f"Model Memory: {self._model.get_memory_footprint()}")
                     return
 
-        if num_gpus > 0 and self._device == "cuda":
+        if num_gpus > 0 and is_hf_accelerate_supported(self._device):
             kwargs.update({"device_map": "auto"})
-        self._model, self._tokenizer = self._load_model(kwargs)
+            is_device_map_auto = True
 
-        if self._device == "mps":
+        self._model, self._tokenizer = self._load_model(**kwargs)
+        self._apply_lora()
+
+        if not is_device_map_auto:
             self._model.to(self._device)
         logger.debug(f"Model Memory: {self._model.get_memory_footprint()}")
 
-    def _select_device(self, device: str) -> str:
-        try:
-            import torch
-        except ImportError:
-            raise ImportError(
-                f"Failed to import module 'torch'. Please make sure 'torch' is installed.\n\n"
-            )
-
-        if device == "auto":
-            # When env CUDA_VISIBLE_DEVICES=-1, torch.cuda.is_available() return False
-            if torch.cuda.is_available():
-                return "cuda"
-            elif torch.backends.mps.is_available():
-                return "mps"
-            return "cpu"
-        elif device == "cuda":
-            if not torch.cuda.is_available():
-                raise ValueError("cuda is unavailable in your environment")
-        elif device == "mps":
-            if not torch.backends.mps.is_available():
-                raise ValueError("mps is unavailable in your environment")
-        elif device == "cpu":
-            pass
-        else:
-            raise ValueError(f"Device {device} is not supported in temporary")
-        return device
-
     @classmethod
     def match(
         cls, llm_family: "LLMFamilyV1", llm_spec: "LLMSpecV1", quantization: str
     ) -> bool:
-        if llm_spec.model_format != "pytorch":
+        if llm_spec.model_format not in ["pytorch", "gptq", "awq"]:
             return False
-        if llm_family.model_name in [
-            "baichuan-chat",
-            "vicuna-v1.3",
-            "falcon",
-            "falcon-instruct",
-            "chatglm",
-            "chatglm2",
-            "chatglm2-32k",
-            "llama-2",
-            "llama-2-chat",
-        ]:
+        model_family = llm_family.model_family or llm_family.model_name
+        if model_family in NON_DEFAULT_MODEL_LIST:
             return False
         if "generate" not in llm_family.model_ability:
             return False
@@ -260,29 +281,34 @@ class PytorchModel(LLM):
     def generate(
         self, prompt: str, generate_config: Optional[PytorchGenerateConfig] = None
     ) -> Union[Completion, Iterator[CompletionChunk]]:
-        from .utils import (
-            generate_stream,
-            generate_stream_chatglm,
-            generate_stream_falcon,
-        )
+        from .utils import generate_stream, generate_stream_falcon
+
+        model_family_name = self.model_family.model_name.lower()
 
         def generator_wrapper(
             prompt: str, generate_config: PytorchGenerateConfig
         ) -> Iterator[CompletionChunk]:
-            if "falcon" in self.model_family.model_name:
-                for completion_chunk, _ in generate_stream_falcon(
-                    self._model, self._tokenizer, prompt, self._device, generate_config
-                ):
-                    yield completion_chunk
-            elif "chatglm" in self.model_family.model_name:
-                for completion_chunk, _ in generate_stream_chatglm(
-                    self._model, self._tokenizer, prompt, self._device, generate_config
+            if "falcon" in model_family_name:
+                for completion_chunk, completion_usage in generate_stream_falcon(
+                    self.model_uid,
+                    self._model,
+                    self._tokenizer,
+                    prompt,
+                    self._device,
+                    generate_config,
                 ):
+                    completion_chunk["usage"] = completion_usage
                     yield completion_chunk
             else:
-                for completion_chunk, _ in generate_stream(
-                    self._model, self._tokenizer, prompt, self._device, generate_config
+                for completion_chunk, completion_usage in generate_stream(
+                    self.model_uid,
+                    self._model,
+                    self._tokenizer,
+                    prompt,
+                    self._device,
+                    generate_config,
                 ):
+                    completion_chunk["usage"] = completion_usage
                     yield completion_chunk
 
         logger.debug(
@@ -294,21 +320,38 @@ class PytorchModel(LLM):
         assert self._model is not None
         assert self._tokenizer is not None
 
+        lora_model = generate_config.pop("lora_name")
+
+        if lora_model is not None and self._peft_model is not None:
+            for lora in self._peft_model:
+                if lora_model == lora.lora_name:
+                    self._model.set_adapter(lora_model)
+                    logger.info(f"Set lora model to {lora_model}")
+                    break
+            else:
+                self._model.disable_adapter()
+                logger.info(f"No lora model {lora_model} found, skip setting")
+
         stream = generate_config.get("stream", False)
         if not stream:
-            if "falcon" in self.model_family.model_name:
+            if "falcon" in model_family_name:
                 for completion_chunk, completion_usage in generate_stream_falcon(
-                    self._model, self._tokenizer, prompt, self._device, generate_config
-                ):
-                    pass
-            elif "chatglm" in self.model_family.model_name:
-                for completion_chunk, completion_usage in generate_stream_chatglm(
-                    self._model, self._tokenizer, prompt, self._device, generate_config
+                    self.model_uid,
+                    self._model,
+                    self._tokenizer,
+                    prompt,
+                    self._device,
+                    generate_config,
                 ):
                     pass
             else:
                 for completion_chunk, completion_usage in generate_stream(
-                    self._model, self._tokenizer, prompt, self._device, generate_config
+                    self.model_uid,
+                    self._model,
+                    self._tokenizer,
+                    prompt,
+                    self._device,
+                    generate_config,
                 ):
                     pass
             completion = Completion(
@@ -338,6 +381,7 @@ class PytorchModel(LLM):
             inputs = input
 
         tokenizer = self._tokenizer
+        tokenizer.pad_token = tokenizer.eos_token
         is_llama = "llama" in str(type(self._model))  # llama supports batch inference
         is_chatglm = "chatglm" in str(type(self._model))
         if is_llama:
@@ -409,6 +453,7 @@ class PytorchChatModel(PytorchModel, ChatModelMixin):
         quantization: str,
         model_path: str,
         pytorch_model_config: Optional[PytorchModelConfig] = None,
+        peft_model: Optional[List[LoRA]] = None,
     ):
         super().__init__(
             model_uid,
@@ -417,50 +462,40 @@ class PytorchChatModel(PytorchModel, ChatModelMixin):
             quantization,
             model_path,
             pytorch_model_config,
+            peft_model,
         )
+        self._context_len = None
 
     def _sanitize_generate_config(
         self,
-        pytorch_generate_config: Optional[PytorchGenerateConfig],
+        generate_config: Optional[PytorchGenerateConfig],
     ) -> PytorchGenerateConfig:
-        pytorch_generate_config = super()._sanitize_generate_config(
-            pytorch_generate_config
-        )
+        generate_config = super()._sanitize_generate_config(generate_config)
         if (
-            "stop" not in pytorch_generate_config
+            (not generate_config.get("stop"))
             and self.model_family.prompt_style
             and self.model_family.prompt_style.stop
         ):
-            pytorch_generate_config["stop"] = self.model_family.prompt_style.stop.copy()
+            generate_config["stop"] = self.model_family.prompt_style.stop.copy()
         if (
-            "stop_token_ids" not in pytorch_generate_config
+            generate_config.get("stop_token_ids", None) is None
             and self.model_family.prompt_style
             and self.model_family.prompt_style.stop_token_ids
         ):
-            pytorch_generate_config[
+            generate_config[
                 "stop_token_ids"
             ] = self.model_family.prompt_style.stop_token_ids.copy()
 
-        return pytorch_generate_config
+        return generate_config
 
     @classmethod
     def match(
         cls, llm_family: "LLMFamilyV1", llm_spec: "LLMSpecV1", quantization: str
     ) -> bool:
-        if llm_spec.model_format != "pytorch":
+        if llm_spec.model_format not in ["pytorch", "gptq", "awq"]:
             return False
-        if llm_family.model_name in [
-            "baichuan-chat",
-            "baichuan-2-chat",
-            "vicuna-v1.3",
-            "falcon",
-            "falcon-instruct",
-            "chatglm",
-            "chatglm2",
-            "chatglm2-32k",
-            "llama-2",
-            "llama-2-chat",
-        ]:
+        model_family = llm_family.model_family or llm_family.model_name
+        if model_family in NON_DEFAULT_MODEL_LIST:
             return False
         if "chat" not in llm_family.model_ability:
             return False
@@ -473,14 +508,21 @@ class PytorchChatModel(PytorchModel, ChatModelMixin):
         chat_history: Optional[List[ChatCompletionMessage]] = None,
         generate_config: Optional[PytorchGenerateConfig] = None,
     ) -> Union[ChatCompletion, Iterator[ChatCompletionChunk]]:
-        assert self.model_family.prompt_style is not None
-        prompt_style = self.model_family.prompt_style.copy()
-        if system_prompt:
-            prompt_style.system_prompt = system_prompt
-        chat_history = chat_history or []
-        full_prompt = self.get_prompt(prompt, chat_history, prompt_style)
+        tools = generate_config.pop("tools", []) if generate_config else None
+        full_prompt = self._get_full_prompt(prompt, system_prompt, chat_history, tools)
 
         generate_config = self._sanitize_generate_config(generate_config)
+        # TODO(codingl2k1): qwen hacky to set stop for function call.
+        model_family = self.model_family.model_family or self.model_family.model_name
+        if tools and model_family in ["qwen-chat", "qwen1.5-chat"]:
+            stop = generate_config.get("stop")
+            if isinstance(stop, str):
+                generate_config["stop"] = [stop, "Observation:"]
+            elif isinstance(stop, Iterable):
+                assert not isinstance(stop, str)
+                generate_config["stop"] = list(stop) + ["Observation:"]
+            else:
+                generate_config["stop"] = "Observation:"
 
         stream = generate_config.get("stream", False)
         if stream:
@@ -490,4 +532,87 @@ class PytorchChatModel(PytorchModel, ChatModelMixin):
         else:
             c = self.generate(full_prompt, generate_config)
             assert not isinstance(c, Iterator)
+            if tools:
+                return self._tool_calls_completion(
+                    self.model_family, self.model_uid, c, tools
+                )
             return self._to_chat_completion(c)
+
+    def load(self):
+        super().load()
+        self._context_len = get_context_length(self._model.config)
+
+    def _get_full_prompt(self, prompt, system_prompt, chat_history, tools):
+        assert self.model_family.prompt_style is not None
+        prompt_style = self.model_family.prompt_style.copy()
+        if system_prompt:
+            prompt_style.system_prompt = system_prompt
+        chat_history = chat_history or []
+        full_prompt = ChatModelMixin.get_prompt(
+            prompt, chat_history, prompt_style, tools=tools
+        )
+        return full_prompt
+
+    def get_max_num_seqs(self) -> int:
+        return self._pytorch_model_config.get("max_num_seqs")  # type: ignore
+
+    def batch_inference(self, req_list: List[InferenceRequest]):
+        from .utils import batch_inference_one_step
+
+        for r in req_list:
+            if r.sanitized_generate_config is None:
+                r.sanitized_generate_config = self._sanitize_generate_config(
+                    r.generate_config
+                )
+            if r.is_prefill:
+                # check some generate params
+                max_src_len = get_max_src_len(self._context_len, r)  # type: ignore
+                if max_src_len < 0:
+                    r.stopped = True
+                    r.error_msg = "Max tokens exceeds model's max length"
+                    continue
+                if r.stream_interval <= 0:
+                    r.stopped = True
+                    r.error_msg = "`stream_interval` must be greater than 0"
+                    continue
+                stop_str = r.sanitized_generate_config.get("stop", None)
+                if stop_str and (
+                    not (isinstance(stop_str, str) or isinstance(stop_str, Iterable))
+                ):
+                    r.stopped = True
+                    r.error_msg = "Invalid `stop` field type"
+                    continue
+                r.full_prompt = self._get_full_prompt(
+                    r.prompt, r.system_prompt, r.chat_history, None
+                )
+
+        assert isinstance(self._context_len, int)
+        batch_inference_one_step(
+            req_list,
+            self.model_uid,
+            self._model,
+            self._tokenizer,
+            self._device,
+            self._context_len,
+        )
+        for req in req_list:
+            if req.stream and req.error_msg is None:
+                if req.completion:
+                    results = []
+                    for i, c in enumerate(req.completion):
+                        if c == "<bos_stream>":
+                            results.append(
+                                self._get_first_chat_completion_chunk(
+                                    req.completion[i + 1]
+                                )
+                            )
+                        elif c == "<eos_stream>":
+                            break
+                        else:
+                            results.append(self._to_chat_completion_chunk(c))
+
+                    if req.stopped and req.include_usage:
+                        results.append(
+                            self._get_final_chat_completion_chunk(req.completion[-1])
+                        )
+                    req.completion = results
